
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:
class Parser
	token DECLARE MODULE CIRCUIT
	token INPUT OUTPUT BIDIRECT INSTRIN INSTROUT
	token INSTRSELF BUS BUS_V SEL SEL_V REG REG_WR REG_WS MEM
	token INSTR_ARG STAGE_NAME TASK INSTRUCT
	token STAGE STATE_NAME SEGMENT_NAME FIRST_STATE STATE SEGMENT
	token PAR ALT ANY IF ELSE
	token GOTO CALL RETURN GENERATE RELAY FINISH
	token NAME NUM CONST
	token OP_ASSIGNREG OP_CONCATENATE OP_SHIFTRIGHT OP_SHIFTLEFT OP_EQEQ OP_FOLDOR OP_FOLDEOR OP_FOLDAND OP_YEN
rule
	sfl_description
		: descriptions_opt
			{ result = Descriptions.new val[0] }

	descriptions_opt
		: # null
		| description descriptions_opt
			{ result = val }

	description
		: module_decl
		| module_def
		| circuit_def

	module_decl
		: DECLARE NAME '{' exterm_defs_opt instr_use_args_opt '}'
			{ result = ModuleDecl.new val[1], val[3], val[4] }

	exterm_defs_opt
		: # null
		| exterm_def exterm_defs_opt
			{ result = val }

	exterm_def
		: INPUT term_list ';'
			{ result = NameList.new :input, val[1] }
		| OUTPUT term_list ';'
			{ result = NameList.new :output, val[1] }
		| BIDIRECT term_list ';'
			{ result = NameList.new :bidirect, val[1] }
		| INSTRIN name_list ';'
			{ result = NameList.new :instrin, val[1] }
		| INSTROUT name_list ';'
			{ result = NameList.new :instrout, val[1] }

	width_opt
		: # null
		| '<' NUM '>'
			{ result = val[1] }

	term_list
		: term term_l_opt
			{ result = val }

	term
		: NAME width_opt
			{ result = val }

	term_l_opt
		: # null
		| ',' term_list
			{ result = val[1] }

	name_list
		: NAME name_l_opt
			{ result = val }

	name_l_opt
		: # null
		| ',' name_list
			{ result = val[1] }

	instr_use_args_opt
		: # null
		| instr_use_arg instr_use_args_opt
			{ result = val }

	instr_use_arg
		: INSTR_ARG NAME '(' name_list_opt ')' ';'
			{ result = InstrArg.new val[1], val[3] }

	name_list_opt
		: # null
		| name_list

	module_def
		: MODULE NAME '{'
		    module_decls_opt  # val[3]
		    term_elem_defs_opt  # val[4]
		    instr_args_opt  # val[5]
		    stage_task_params_opt  # val[6]
		    common_action_opt  # val[7], common (always) action
		    instructs_opt  # val[8]
		    stages_opt  # val[9]
		  '}'
			{ result = ModuleDef.new val[1], val[3], val[4], val[5], val[6], val[7], val[8], val[9] }

	circuit_def
		: CIRCUIT NAME '{'
		    module_decls_opt  # val[3]
		    term_elem_defs_opt  # val[4]
		    instr_args_opt  # val[5]
		    stage_task_params_opt  # val[6]
		    common_action_opt  # val[7], common (always) action
		    instructs_opt  # val[8]
		    stages_opt  # val[9]
		  '}'
			{ result = CircuitDef.new val[1], val[3], val[4], val[5], val[6], val[7], val[8], val[9] }

	module_decls_opt
		: # null
		| module_decl module_decls_opt
			{ result = val }

	term_elem_defs_opt
		: # null
		| term_elem_def term_elem_defs_opt
			{ result = val }

	term_elem_def
		: exterm_def
		| elem_def

	elem_def
		: INSTRSELF name_list ';'
			{ result = NameList.new :instrself, val[1] }
		| BUS term_list ';'
			{ result = NameList.new :bus, val[1] }
		| SEL term_list ';'
			{ result = NameList.new :sel, val[1] }
		| BUS_V term_list ';'
			{ result = NameList.new :bus_v, val[1] }
		| SEL_V term_list ';'
			{ result = NameList.new :sel_v, val[1] }
		| REG term_list ';'
			{ result = NameList.new :reg, val[1] }
		| REG_WS term_list ';'
			{ result = NameList.new :reg_ws, val[1] }
		| REG_WR term_list ';'
			{ result = NameList.new :reg_wr, val[1] }
		| MEM mem_list ';'
			{ result = NameList.new :mem, val[1] }
		| NAME name_list ';'  # submodule
			{ result = NameList.new val[0], val[1] }

	mem_list
		: mem mem_opt
			{ result = val }

	mem
		: NAME '[' NUM ']' width_opt
			{ result = [val[0], val[2], val[4]] }

	mem_opt
		: # null
		| ',' mem_list
			{ result = val[1] }

	instr_args_opt
		: # null
		| instr_arg instr_args_opt
			{ result = val }

	instr_arg
		: INSTR_ARG NAME '(' name_list_opt ')' ';'
			{ result = InstrArg.new val[1], val[3] }
		| INSTR_ARG NAME '.' NAME '(' name_list_opt ')' ';'
			{ result = InstrArgEx.new val[1], val[3], val[5] }

	stage_task_params_opt
		: # null
		| stage_task_param stage_task_params_opt
			{ result = val }

	stage_task_param
		: STAGE_NAME NAME '{' tasks '}'
			{ result = StageName.new val[1], val[3] }

	tasks
		: task tasks_opt
			{ result = val }

	tasks_opt
		: # null
		| tasks

	task
		: TASK NAME '(' name_list_opt ')' ';'
			{ result = Task.new val[1], val[3] }

	common_action_opt
		: # null
		| action

	instructs_opt
		: # null
		| instruct instructs_opt
			{ result = val }

	instruct
		: INSTRUCT NAME action
			{ result = Instruct.new val[1], val[2] }
		| INSTRUCT NAME '.' NAME action
			{ result = InstructEx.new val[1], val[3], val[4] }

	stages_opt
		: # null
		| stage stages_opt
			{ result = val }

	stage
		: STAGE NAME '{'
		    state_names_opt  # val[3]
		    segment_names_opt  # val[4]
		    first_state_opt  # val[5]
		    common_action_opt  # val[6]
		    states_opt  # val[7]
		    segments_opt  # val[8]
		  '}'
			{ result = Stage.new val[1], val[3], val[4], val[5], val[6], val[7], val[8] }

	segments_opt
		: # null
		| segment segments_opt
			{ result = val }

	segment
		: SEGMENT NAME '{'
		    state_names  # val[3]
		    first_state  # val[4]
		    common_action_opt  # val[5]
		    states  # val[6]
		  '}'
			{ result = Segment.new val[1], val[3], val[4], val[5], val[6] }

	state_names_opt
		: # null
		| state_names

	state_names
		: state_name state_names_opt
			{ result = val }

	state_name
		: STATE_NAME name_list ';'
			{ result = NameList.new :state_name, val[1] }

	segment_names_opt
		: # null
		| segment_names

	segment_names
		: segment_name segment_names_opt
			{ result = val }

	segment_name
		: SEGMENT_NAME name_list ';'
			{ result = NameList.new :segment_name, val[1] }

	first_state_opt
		: # null
		| first_state

	first_state
		: FIRST_STATE NAME ';'
			{ result = val[1] }

	states_opt
		: # null
		| states

	states
		: state states_opt
			{ result = val }

	state
		: STATE NAME action
			{ result = State.new val[1], val[2] }

	action
		: PAR '{' actions_opt '}'
			{ result = ActPar.new val[2] }
		| ALT '{' guarded_actions else_action_opt '}'
			{ result = ActAlt.new val[2], val[3] }
		| ANY '{' guarded_actions else_action_opt '}'
			{ result = ActAny.new val[2], val[3] }
		| IF '(' exp ')' action
			{ result = ActIf.new val[2], val[4] }
		| uaction

	actions_opt
		: # null
		| action actions_opt
			{ result = val }

	guarded_actions
		: guarded_action guarded_actions_opt
			{ result = val }

	guarded_actions_opt
		: # null
		| guarded_action guarded_actions_opt
			{ result = val }

	guarded_action
		: exp ':' action
			{ result = GuardedAct.new val[0], val[2] }

	else_action_opt
		: # null
		| ELSE ':' action
			{ result = val[2] }

	uaction
		: NAME '=' exp ';'
			{ result = UActOutput.new val[0], val[2] }
		| NAME '.' NAME '=' exp ';'
			{ result = UActOutputEx.new val[0], val[2], val[4] }
		| NAME OP_ASSIGNREG exp ';'  # :=
			{ result = UActRegWrite.new val[0], val[2] }
		| NAME '[' exp ']' OP_ASSIGNREG exp ';'  # :=
			{ result = UActMemWrite.new val[0], val[2], val[5] }
		| GOTO NAME ';'
			{ result = UActGoto.new val[1] }
		| CALL NAME '(' name_opt ')' ';'
			{ result = UActCall.new val[1], val[3] }
		| RETURN ';'
			{ result = UActReturn.new }
		| GENERATE NAME '.' NAME '(' exp_list_opt ')' ';'
			{ result = UActGenerate.new val[1], val[3], val[5] }
		| RELAY NAME '.' NAME '(' exp_list_opt ')' ';'
			{ result = UActRelay.new val[1], val[3], val[5] }
		| FINISH ';'
			{ result = UActFinish.new }
		| NAME '(' exp_list_opt ')' ';'
			{ result = UActInstruct.new val[0], val[2] }
		| NAME '.' NAME '(' exp_list_opt ')' ';'
			{ result = UActInstructEx.new val[0], val[2], val[4] }
		| ';'  # nop
			{ result = UActNop.new }

	name_opt
		: # null
		| NAME

	exp_list_opt
		: # null
		| exp_list

	exp_list
		: exp exp_l_opt

	exp_l_opt
		: # null
		| ',' exp_list

	exp
		: unary binary_op exp
			{ result = BExp.new val[0], val[1].intern, val[2] }
		| unary

	binary_op
		: '|' | '@' | '&' | OP_CONCATENATE | '+' | OP_SHIFTRIGHT | OP_SHIFTLEFT | OP_EQEQ

	unary
		: unary_op unary
			{ result = UExp.new val[0].intern, val[1] }
		| NUM '#' unary
			{ result = SignExt.new val[0], val[2] }
		| elem '<' NUM '>'
			{ result = CutOut.new val[0], val[2] }
		| elem '<' NUM ':' NUM '>'
			{ result = CutOut2.new val[0], val[2], val[4] }
		| elem

	unary_op
		: '^' | OP_FOLDOR | OP_FOLDEOR | OP_FOLDAND | '/' | OP_YEN

	elem
		: '(' exp ')'
			{ result = ElemParen.new val[1] }
		| CONST
		| NAME
		| NAME '[' exp ']'
			{ result = ElemMemRead.new val[0], val[2] }
		| NAME '.' NAME
			{ result = ElemEx.new val[0], val[2] }
		| NAME '(' exp_list_opt ')' '.' NAME
			{ result = ElemInstruct.new val[0], val[2], val[5] }
		| NAME '.' NAME '(' exp_list_opt ')' '.' NAME
			{ result = ElemInstructEx.new val[0], val[2], val[4], val[7] }
end

---- header

require 'pp'

require 'sfllib'

module SFL

---- inner

def parse lex
	@lex = lex
	do_parse
end

def next_token
	@lex.next_token
end

def on_error t, val, vstack
	raise ParseError, "\n\nunexpected token #{val.inspect} in line #{@lex.line_counter}\n\n" # +
	#	"t == #{t.inspect}\n" +
	#	"vstack == #{vstack.inspect}\n"
end

---- footer

end
