
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:
module SFL
	def SFL.flatten tree
		l = []
		while tree
			head, tree = tree
			l << head
		end
		l
	end

	def SFL.define_class class_name, *attrs
		l = []
		attrs.each do |a|
			if /\A\*/ =~ a then
				l << a[1, a.size - 1]
			else
				l << a
			end
		end
		args = l.join ', '
		settings = ''
		attrs.each do |a|
			if /\A\*/ =~ a then
				a = a[1, a.size - 1]
				settings << "\t\t@#{a} = SFL.flatten #{a}\n"
			else
				settings << "\t\t@#{a} = #{a}\n"
			end
		end
		defstring = "class #{class_name}\n\tdef initialize #{args}\n#{settings}\tend\nend\n"
		#print defstring
		SFL.module_eval defstring
	end

	SFL.define_class 'Descriptions', '*descriptions'
	SFL.define_class 'ModuleDecl', 'name', '*exterms', '*instr_use_args'

	class NameList
		def initialize type, list
			@type = type
			l = SFL.flatten list
			@list = l.map do|name, *opt|
				if opt and not opt.empty? then
					NameEntry.new name, opt
				else
					NameEntry.new name, nil
				end
			end
		end
	end

	SFL.define_class 'NameEntry', 'name', 'opt'

	class NameEntry
		attr_reader :name, :opt
	end

	class InstrArg
		def initialize name, list
			@name = name
			l = SFL.flatten list
			@list = l.map {|name| NameEntry.new name, nil }
		end
	end

	class InstrArgEx
		def initialize mod_name, name, list
			@mod_name = mod_name
			@name = name
			l = SFL.flatten list
			@list = l.map {|name| NameEntry.new name, nil }
		end
	end

	SFL.define_class 'ModuleDef', 'name', '*module_decls', '*term_elem_defs', '*instr_args', '*stage_names', 'common_action', '*instructs', '*stages'
	SFL.define_class 'CircuitDef', 'name', '*module_decls', '*term_elem_defs', '*instr_args', '*stage_names', 'common_action', '*instructs', '*stages'
	SFL.define_class 'StageName', 'name', '*tasks'

	class Task
		def initialize name, list
			@name = name
			l = SFL.flatten list
			@list = l.map {|name| NameEntry.new name, nil }
		end
	end

	SFL.define_class 'Instruct', 'name', 'action'
	SFL.define_class 'InstructEx', 'mod_name', 'name', 'action'
	SFL.define_class 'Stage', 'name', '*state_names', '*segment_names', 'first_state', 'common_action', '*states', '*segments'
	SFL.define_class 'Segment', 'name', '*state_names', 'first_state', 'common_action', '*states'
	SFL.define_class 'State', 'name', 'action'
	SFL.define_class 'ActPar', '*actions'
	SFL.define_class 'ActAlt', '*gurded_actions', 'else_action'
	SFL.define_class 'ActAny', '*gurded_actions', 'else_action'
	SFL.define_class 'ActIf', 'exp', 'action'
	SFL.define_class 'GuardedAct', 'exp', 'action'
	SFL.define_class 'UActOutput', 'name', 'exp'
	SFL.define_class 'UActOutputEx', 'mod_name', 'name', 'exp'
	SFL.define_class 'UActRegWrite', 'name', 'exp'
	SFL.define_class 'UActMemWrite', 'name', 'adr_exp', 'exp'
	SFL.define_class 'UActGoto', 'name'
	SFL.define_class 'UActCall', 'seg_name', 'state_name'
	SFL.define_class 'UActReturn'
	SFL.define_class 'UActGenerate', 'stage_name', 'task_name', '*exps'
	SFL.define_class 'UActRelay', 'stage_name', 'task_name', '*exps'
	SFL.define_class 'UActFinish'
	SFL.define_class 'UActInstruct', 'name', '*exps'
	SFL.define_class 'UActInstructEx', 'mod_name', 'name', '*exps'
	SFL.define_class 'UActNop'
	SFL.define_class 'BExp', 'unary', 'op', 'exp'
	SFL.define_class 'UExp', 'op', 'unary'
	SFL.define_class 'SignExt', 'num', 'unary'
	SFL.define_class 'CutOut', 'elem', 'num'
	SFL.define_class 'CutOut2', 'elem', 'msb_num', 'lsb_num'
	SFL.define_class 'ElemParen', 'exp'
	SFL.define_class 'ElemMemRead', 'name', 'exp'
	SFL.define_class 'ElemEx', 'mod_name', 'name'
	SFL.define_class 'ElemInstruct', 'instr_name', '*exps', 'term_name'
	SFL.define_class 'ElemInstructEx', 'mod_name', 'instr_name', '*exps', 'term_name'
end
